import requests
from ..urls import voice_list_url
from ..token import Token


def format_voice(voice: dict) -> str:
    return f"""{voice["Name"]}
Display Name: {voice["DisplayName"]}
Local Name: {voice["LocalName"]} @ {voice["Locale"]}
Locale: {voice["LocaleName"]}
Gender: {voice["Gender"]}
ID: {voice["ShortName"]}
Styles: {voice.get("StyleList")}
Voice Type: {voice["VoiceType"]}
Status: {voice["Status"]}
"""


def get_voice_list(token: Token) -> list:
    response = requests.get(voice_list_url(token.region),
                            headers={'Authorization': 'Bearer ' + token.token})
    return response.json()


def list_voices(args):
    voices = get_voice_list(Token())
    if hasattr(args, 'voice'):
        voices = [v for v in voices if v["ShortName"] == args.voice]
    if hasattr(args, 'locale'):
        voices = [v for v in voices if v['Locale'] == args.locale]
    for voice in voices:
        print(format_voice(voice))



import azure.cognitiveservices.speech as speechsdk

from ..token import Token


class SpeechServiceProvider:
    """
    The SpeechServiceProvider class is a service provider for Azure Cognitive Services Text-to-speech
    that automatically renews trial auth tokens.
    """

    def __init__(self):
        self._current_token = Token()

    @property
    def token(self) -> Token:
        """
        Returns the current valid token instance.
        :return: token instance of type Token
        """
        if self._expired:
            self.renew()
        return self._current_token

    @property
    def _expired(self) -> bool:
        return self._current_token.expired()

    def renew(self) -> None:
        """
        Manually renew the current token. Usually you do not need to call this method.
        :return: None
        """
        self._current_token.renew()

    def text_to_speech(self, text: str, cfg: speechsdk.SpeechConfig,
                       output: speechsdk.audio.AudioOutputConfig) -> speechsdk.SpeechSynthesisResult:
        if self._expired:
            self.renew()
        return speechsdk.SpeechSynthesizer(speech_config=cfg, audio_config=output).speak_text(text)

    def ssml_to_speech(self, ssml: str, cfg: speechsdk.SpeechConfig,
                       output: speechsdk.audio.AudioOutputConfig) -> speechsdk.SpeechSynthesisResult:
        if self._expired:
            self.renew()
        return speechsdk.SpeechSynthesizer(speech_config=cfg, audio_config=output).speak_ssml(ssml)



from typing import Union
from xml.sax.saxutils import escape


# pylint: disable=too-many-arguments
def create_ssml(text: str, voice: Union[str, None], rate: Union[float, str], pitch: Union[float, str],
                style: str = "general", style_degree: Union[float, None] = None, role: Union[str, None] = None) -> str:
    """
    Create SSML for text to be spoken.
    """
    style_degree_text = f'styledegree="{style_degree}"' if style_degree is not None else ""
    role_text = f'role="{role}"' if role is not None else ""
    if isinstance(rate, str):
        rate_text = f'rate="{rate}"'
    else:
        rate_text = f'rate="{round(rate * 100, 2)}%"'
    if isinstance(pitch, str):
        pitch_text = f'pitch="{pitch}"'
    else:
        pitch_text = f'pitch="{round(pitch * 100, 2)}%"'
    ssml = '<speak xmlns="http://www.w3.org/2001/10/synthesis" xmlns:mstts="http://www.w3.org/2001/mstts" ' \
           'xmlns:emo="http://www.w3.org/2009/10/emotionml" version="1.0" xml:lang="en-US"> '
    ssml += f'<voice name="{voice}">' if voice is not None else '<voice>'
    ssml += f'<mstts:express-as style="{style}" {style_degree_text} {role_text}>'
    ssml += f'<prosody {rate_text} {pitch_text}>'
    ssml += escape(text)
    ssml += '</prosody></mstts:express-as></voice></speak>'
    return ssml



class Range:
    # pylint: disable=line-too-long
    """
    A range of values for argparse.
    reference: https://stackoverflow.com/questions/12116685/how-can-i-require-my-python-scripts-argument-to-be-a-float-in-a-range-using-arg
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __eq__(self, other):
        return self.start <= other <= self.end

    def __repr__(self):
        return f'values in range {self.start}-{self.end} (inclusive)'



import sys
from functools import partial

import azure.cognitiveservices.speech as speechsdk

from ..api import SpeechServiceProvider, text_to_speech, pure_text_to_speech, ssml_to_speech
from ..ssml import create_ssml
from ..errors import AspeakError
from .voices import list_voices
from .utils import list_qualities_and_formats
from ..quality import QUALITIES
from .parser import parser
from .constants import COLOR_RED, COLOR_CLEAR
from .validation import validate_quality, get_ineffective_args_for_listing, has_text_options
from .result_handler import handle_result


def read_file(args):
    if not hasattr(args, 'file') or args.file == '-':
        return sys.stdin.read()
    with open(args.file, 'r', encoding=args.encoding) as file:
        return file.read()


def preprocess_text(text, args):
    """
    Preprocess text.
    :param text: plain text
    :param args: args
    :return: (is_ssml, text_or_ssml, extra_options)
    """
    if has_text_options(args):
        if args.voice is None:
            parser.error('Voice must be specified when using options for --text')
        pitch = args.pitch if hasattr(args, 'pitch') else 0.0
        rate = args.rate if hasattr(args, 'rate') else 0.0
        voice = args.voice
        style = args.style if hasattr(args, 'style') else 'general'
        role = args.role if hasattr(args, 'role') else None
        style_degree = args.style_degree if hasattr(args, 'style_degree') else None
        ssml = create_ssml(text, voice, rate, pitch, style, style_degree, role)
        return True, ssml, None
    return False, text, {'locale': args.locale, 'voice': args.voice}


def speech_function_selector(funcs, preprocessed, audio_format):
    _ssml_to_speech, _pure_text_to_speech = funcs
    is_ssml, text_or_ssml, options = preprocessed
    if is_ssml:
        return _ssml_to_speech(text_or_ssml, audio_format)
    return _pure_text_to_speech(text_or_ssml, audio_format=audio_format, **options)


def main_ssml(func, args, audio_format):
    if hasattr(args, 'rate') or hasattr(args, 'pitch') or hasattr(args, 'style'):
        parser.error(
            'You can only use text options with --text. Please set these settings in your SSML.')
    if args.ssml is None:
        # --ssml is provided but empty
        result = func(read_file(args), audio_format)
    else:
        # --ssml is provided and not empty
        if hasattr(args, 'file'):
            parser.error('You can only specify one input source.')
        result = func(args.text, audio_format)
    return result


def main_text(funcs, args, audio_format):
    if args.text is None:
        # --text is provided but empty
        result = speech_function_selector(funcs, preprocess_text(read_file(args), args), audio_format)
    else:
        # --text is provided and not empty
        if hasattr(args, 'file'):
            parser.error('You can only specify one input source.')
        result = speech_function_selector(funcs, preprocess_text(args.text, args), audio_format)
    return result


def main_list_qualities_and_formats(args):
    ineffective_args = get_ineffective_args_for_listing(args)
    if hasattr(args, 'locale'):
        parser.error('--locale can not be used with --list-qualities-and-formats')
    if hasattr(args, 'voice'):
        parser.error('--voice can not be used with --list-qualities-and-formats')
    if ineffective_args:
        parser.error(f"You can't use argument(s) {ineffective_args} with --list-qualities-and-formats.")
    list_qualities_and_formats()


def main_list_voices(args):
    ineffective_args = get_ineffective_args_for_listing(args)
    if ineffective_args:
        parser.error(f"You can't use argument(s) {ineffective_args} with --list-voices.")
    list_voices(args)


def setup_defaults(args):
    if args.output_path is None:
        args.audio_config = speechsdk.audio.AudioOutputConfig(use_default_speaker=True)
    else:
        args.audio_config = speechsdk.audio.AudioOutputConfig(filename=args.output_path)

    args.locale = args.locale if hasattr(args, 'locale') else None
    args.voice = args.voice if hasattr(args, 'voice') else None
    args.quality = args.quality if hasattr(args, 'quality') else 0
    args.encoding = args.encoding if hasattr(args, 'encoding') else 'utf-8'

    args.file_type = "wav"  # The output file format
    for ext in ["mp3", "ogg", "webm"]:
        # mp3, ogg, webm are only supported when outputting to file
        if args.output_path is None and getattr(args, ext):
            parser.error(f"{ext} format is only supported when outputting to a file.")
        if getattr(args, ext):
            args.file_type = ext
    if args.file_type == "wav":
        # Set --wav to true in case that no format argument is provided
        args.wav = True


def main():
    args = parser.parse_args()

    if args.list_qualities_and_formats:
        main_list_qualities_and_formats(args)
        return

    if args.list_voices:
        main_list_voices(args)
        return

    setup_defaults(args)
    validate_quality(args, parser)

    if hasattr(args, 'format'):
        audio_format = getattr(speechsdk.SpeechSynthesisOutputFormat, args.format)
    else:
        audio_format = QUALITIES[args.file_type][args.quality]

    try:
        provider = SpeechServiceProvider()
        _text_to_speech = partial(text_to_speech, provider, args.audio_config)
        _ssml_to_speech = partial(ssml_to_speech, provider, args.audio_config)
        _pure_text_to_speech = partial(pure_text_to_speech, provider, args.audio_config)
        funcs = _ssml_to_speech, _pure_text_to_speech
        if hasattr(args, 'ssml'):
            result = main_ssml(_ssml_to_speech, args, audio_format)
        elif hasattr(args, 'text'):
            result = main_text(funcs, args, audio_format)
        else:
            # Neither --text nor --ssml is provided, pretend --text is provided and empty
            result = speech_function_selector(funcs, preprocess_text(read_file(args), args), audio_format)
        handle_result(result)
    except AspeakError as exception:
        print(f"{COLOR_RED}Error{COLOR_CLEAR}: {exception}")
        sys.exit(4)



class AspeakError(Exception):
    """
    Base class for all Aspeak errors.
    """


class TokenRetrievalError(AspeakError):
    """
    Error raised when the trial token cannot be retrieved.
    """

    def __init__(self, status_code, message="Failed to retrieve token"):
        super().__init__(message)
        self.status_code = status_code
        self.message = message



import argparse
from ..version import version
from .range import Range
# pylint: disable=redefined-builtin
from .value_parsers import pitch, rate, format

parser = argparse.ArgumentParser(
    description='Try speech synthesis service(Provided by Azure Cognitive Services) in your terminal!',
    prog='aspeak')
group = parser.add_mutually_exclusive_group()
group.add_argument('-V', '--version', action='version', version=f'%(prog)s {version}')
group.add_argument('-L', '--list-voices', action='store_true',
                   help='list available voices, you can combine this argument with -v and -l', dest='list_voices')
group.add_argument('-Q', '--list-qualities-and-formats', action='store_true',
                   help='list available qualities and formats',
                   dest='list_qualities_and_formats')
subgroup = group.add_mutually_exclusive_group()
subgroup.add_argument('-t', '--text', help='Text to speak. Left blank when reading from file/stdin',
                      dest='text', nargs='?', default=argparse.SUPPRESS)
subgroup.add_argument('-s', '--ssml', help='SSML to speak. Left blank when reading from file/stdin',
                      dest='ssml', nargs='?', default=argparse.SUPPRESS)
text_group = parser.add_argument_group('Options for --text')
text_group.add_argument('-p', '--pitch', help='Set pitch, default to 0. Valid values include floats(will be converted '
                                              'to percentages), percentages such as 20%% and -10%%, '
                                              'absolute values like 300Hz, and relative values like -20Hz, +2st '
                                              'and string values like x-low. See the documentation for more details.',
                        dest='pitch', type=pitch, default=argparse.SUPPRESS)
text_group.add_argument('-r', '--rate', help='Set speech rate, default to 0. Valid values include floats(will be '
                                             'converted to percentages), percentages like -20%%, floats with postfix '
                                             '"f" (e.g. 2f means doubling the default speech rate), and string '
                                             'values like x-slow. See the documentation for more details.',
                        dest='rate', type=rate, default=argparse.SUPPRESS)
text_group.add_argument('-S', '--style', help='Set speech style, default to "general"', dest='style',
                        default=argparse.SUPPRESS)
text_group.add_argument('-R', '--role',
                        help='Specifies the speaking role-play. This only works for some Chinese voices!',
                        dest='role', type=str, default=argparse.SUPPRESS,
                        choices=['Girl', 'Boy', 'YoungAdultFemale', 'YoungAdultMale', 'OlderAdultFemale',
                                 'OlderAdultMale', 'SeniorFemale', 'SeniorMale'])
text_group.add_argument('-d', '--style-degree', dest='style_degree', type=float, default=argparse.SUPPRESS,
                        help='Specifies the intensity of the speaking style.'
                             'This only works for some Chinese voices!', choices=[Range(0.01, 2)])
parser.add_argument('-f', '--file', help='Text/SSML file to speak, default to `-`(stdin)', dest='file',
                    default=argparse.SUPPRESS)
parser.add_argument('-e', '--encoding', help='Text/SSML file encoding, default to "utf-8"(Not for stdin!)',
                    dest='encoding', default=argparse.SUPPRESS)
parser.add_argument('-o', '--output', help='Output file path, wav format by default', dest='output_path', default=None)
format_group = parser.add_mutually_exclusive_group()
format_group.add_argument('--mp3', help='Use mp3 format for output. (Only works when outputting to a file)',
                          action='store_true', dest='mp3')
format_group.add_argument('--ogg', help='Use ogg format for output. (Only works when outputting to a file)',
                          action='store_true', dest='ogg')
format_group.add_argument('--webm', help='Use webm format for output. (Only works when outputting to a file)',
                          action='store_true', dest='webm')
format_group.add_argument('--wav', help='Use wav format for output', action='store_true', dest='wav')
format_group.add_argument('-F', '--format', help='Set output audio format (experts only)', dest='format', type=format,
                          default=argparse.SUPPRESS)
parser.add_argument('-l', '--locale', help='Locale to use, default to en-US', dest='locale', default=argparse.SUPPRESS)
parser.add_argument('-v', '--voice', help='Voice to use', dest='voice', default=argparse.SUPPRESS)
parser.add_argument('-q', '--quality', help='Output quality, default to 0', dest='quality', type=int,
                    default=argparse.SUPPRESS)
# pylint: disable=line-too-long
parser.usage = '''aspeak [-h] [-V | -L | -Q | [-t [TEXT] [-p PITCH] [-r RATE] [-S STYLE] [-R ROLE] [-d STYLE_DEGREE] | -s [SSML]]]
              [-f FILE] [-e ENCODING] [-o OUTPUT_PATH] [-l LOCALE] [-v VOICE]
              [--mp3 [-q QUALITY] | --ogg [-q QUALITY] | --webm [-q QUALITY] | --wav [-q QUALITY] | -F FORMAT] 
'''
parser.epilog = 'Attention: If the result audio is longer than 10 minutes, the audio will be truncated to 10 minutes ' \
                'and the program will not report an error. Unreasonable high/low values for pitch and rate will be ' \
                'clipped to reasonable values by Azure Cognitive Services.' \
                'Please refer to the documentation for other limitations at' \
                ' https://github.com/kxxt/aspeak/blob/main/README.md#limitations. By the way, we don\'t store your ' \
                'data, and Microsoft doesn\'t store your data according to information available on ' \
                'https://azure.microsoft.com/en-us/services/cognitive-services/text-to-speech/'



from time import time

from .auth import _parse_info_from_token, _get_auth_token


class Token:
    def __init__(self, token=None):
        if token is None:
            token = _get_auth_token()
        self.token = token
        info = _parse_info_from_token(token)
        self.region = info['region']
        self.expires = info['exp']

    def expired(self):
        # 5 seconds safe margin
        return self.expires < time() - 5

    def renew(self):
        self.__init__(_get_auth_token())

    def __repr__(self):
        return f'Token({self.token})'

    @classmethod
    def from_string(cls, token):
        return cls(token)



from azure.cognitiveservices.speech import SpeechSynthesisOutputFormat

QUALITIES = {
    'wav': {
        -2: SpeechSynthesisOutputFormat.Riff8Khz16BitMonoPcm,
        -1: SpeechSynthesisOutputFormat.Riff16Khz16BitMonoPcm,
        0: SpeechSynthesisOutputFormat.Riff24Khz16BitMonoPcm,
        1: SpeechSynthesisOutputFormat.Riff24Khz16BitMonoPcm,
    },
    'mp3': {
        -4: SpeechSynthesisOutputFormat.Audio16Khz32KBitRateMonoMp3,
        -3: SpeechSynthesisOutputFormat.Audio16Khz64KBitRateMonoMp3,
        -2: SpeechSynthesisOutputFormat.Audio16Khz128KBitRateMonoMp3,
        -1: SpeechSynthesisOutputFormat.Audio24Khz48KBitRateMonoMp3,
        0: SpeechSynthesisOutputFormat.Audio24Khz96KBitRateMonoMp3,
        1: SpeechSynthesisOutputFormat.Audio24Khz160KBitRateMonoMp3,
        2: SpeechSynthesisOutputFormat.Audio48Khz96KBitRateMonoMp3,
        3: SpeechSynthesisOutputFormat.Audio48Khz192KBitRateMonoMp3,
    },
    'ogg': {
        -1: SpeechSynthesisOutputFormat.Ogg16Khz16BitMonoOpus,
        0: SpeechSynthesisOutputFormat.Ogg24Khz16BitMonoOpus,
        1: SpeechSynthesisOutputFormat.Ogg48Khz16BitMonoOpus,
    },
    'webm': {
        -1: SpeechSynthesisOutputFormat.Webm16Khz16BitMonoOpus,
        0: SpeechSynthesisOutputFormat.Webm24Khz16BitMonoOpus,
        1: SpeechSynthesisOutputFormat.Webm24Khz16Bit24KbpsMonoOpus,
    }
}



TRAIL_URL = 'https://azure.microsoft.com/en-us/services/cognitive-services/text-to-speech/'


def voice_list_url(region: str) -> str:
    return f'https://{region}.tts.speech.microsoft.com/cognitiveservices/voices/list'



from typing import Union

import azure.cognitiveservices.speech as speechsdk

from .provider import SpeechServiceProvider
from .format import AudioFormat, FileFormat, parse_format
from ..ssml import create_ssml


# pylint: disable=too-many-arguments
def pure_text_to_speech(provider: SpeechServiceProvider, output: speechsdk.audio.AudioOutputConfig, text: str,
                        locale: Union[str, None] = None, voice: Union[str, None] = None,
                        audio_format: Union[
                            AudioFormat, FileFormat, speechsdk.SpeechSynthesisOutputFormat, None] = None) \
        -> speechsdk.SpeechSynthesisResult:
    """
    Execute a text-to-speech request without SSML.
    :param provider: An instance of SpeechServiceProvider.
    :param output: An instance of AudioOutputConfig.
    :param text: The text to be synthesized.
    :param locale: The locale of the voice, optional.
    :param voice: The voice name, optional.
    :param audio_format: The audio format, optional.
    :return: result of type SpeechSynthesisResult.
    """
    cfg = speechsdk.SpeechConfig(auth_token=provider.token.token, region=provider.token.region)
    if locale is not None:
        cfg.speech_synthesis_language = locale
    if voice is not None:
        cfg.speech_synthesis_voice_name = voice
    cfg.set_speech_synthesis_output_format(parse_format(audio_format))
    return provider.text_to_speech(text, cfg, output)


# pylint: disable=too-many-arguments
def text_to_speech(provider: SpeechServiceProvider, output: speechsdk.audio.AudioOutputConfig, text: str, voice: str,
                   rate: Union[str, float] = 0.0, pitch: Union[str, float] = 0.0, style: str = "general",
                   style_degree: Union[float, None] = None,
                   role: Union[str, None] = None,
                   audio_format: Union[AudioFormat, FileFormat, speechsdk.SpeechSynthesisOutputFormat, None] = None) \
        -> speechsdk.SpeechSynthesisResult:
    """
    Execute a text-to-speech request with generated SSML from text and various options.
    :param provider: An instance of SpeechServiceProvider.
    :param output: An instance of AudioOutputConfig.
    :param text: The text to be synthesized.
    :param voice: The voice name.
    :param rate: The speaking rate, optional. See the documentation for more details.
    :param pitch: The speaking pitch, optional. See the documentation for more details.
    :param style: The speaking style, optional. See the documentation for more details.
    :param style_degree: The speaking style degree, optional. It only works for some Chinese voices.
    :param role: The speaking role, optional. It only works for some Chinese voices.
    :param audio_format: The audio format, optional.
    :return: result of type SpeechSynthesisResult.
    """
    ssml = create_ssml(text, voice, rate, pitch, style, style_degree, role)
    cfg = speechsdk.SpeechConfig(auth_token=provider.token.token, region=provider.token.region)
    cfg.set_speech_synthesis_output_format(parse_format(audio_format))
    return provider.ssml_to_speech(ssml, cfg, output)


def ssml_to_speech(provider: SpeechServiceProvider, output: speechsdk.audio.AudioOutputConfig, ssml: str,
                   audio_format: Union[AudioFormat, FileFormat, speechsdk.SpeechSynthesisOutputFormat, None]) \
        -> speechsdk.SpeechSynthesisResult:
    """
    Execute a text-to-speech request with SSML.
    :param provider: An instance of SpeechServiceProvider.
    :param output: An instance of AudioOutputConfig.
    :param ssml: The SSML to be synthesized.
    :param audio_format: The audio format, optional.
    :return: result of type SpeechSynthesisResult.
    """
    cfg = speechsdk.SpeechConfig(auth_token=provider.token.token, region=provider.token.region)
    cfg.set_speech_synthesis_output_format(parse_format(audio_format))
    return provider.ssml_to_speech(ssml, cfg, output)



from enum import Enum
from typing import Union

import azure.cognitiveservices.speech as speechsdk

from ..quality import QUALITIES


class FileFormat(Enum):
    """
    Enum for audio file formats.
    """
    WAV = 'wav'
    MP3 = 'mp3'
    OGG = 'ogg'
    WEBM = 'webm'


# pylint: disable=too-few-public-methods
class AudioFormat:
    """
    Class for audio format. You can also use FileFormat enum or speechsdk.SpeechSynthesisOutputFormat.
    """

    def __init__(self, file_format: FileFormat, quality: int = 0):
        """
        Initialize an instance of AudioFormat from the file format and quality.
        :param file_format: Enum of type FileFormat
        :param quality: Quality of the audio, execute `aspeak -Q` to see the available qualities for each file format.
        """
        self._format = QUALITIES[file_format.value][quality]

    @property
    def format(self) -> speechsdk.SpeechSynthesisOutputFormat:
        """
        Get the underlying format.
        :return: audio format of type speechsdk.SpeechSynthesisOutputFormat
        """
        return self._format


def parse_format(audio_format: Union[AudioFormat, FileFormat, speechsdk.SpeechSynthesisOutputFormat, None]) \
        -> speechsdk.SpeechSynthesisOutputFormat:
    if isinstance(audio_format, AudioFormat):
        return audio_format.format
    if isinstance(audio_format, FileFormat):
        return QUALITIES[audio_format.value][0]
    if isinstance(audio_format, speechsdk.SpeechSynthesisOutputFormat):
        return audio_format
    if audio_format is None:
        return QUALITIES['wav'][0]
    raise ValueError(f'Invalid audio format: {audio_format}')



import sys

import azure.cognitiveservices.speech as speechsdk

from .constants import COLOR_CLEAR, COLOR_RED


def handle_result(result: speechsdk.SpeechSynthesisResult):
    if result.reason == speechsdk.ResultReason.SynthesizingAudioCompleted:
        sys.exit(0)
    elif result.reason == speechsdk.ResultReason.Canceled:
        cancellation_details = result.cancellation_details
        print(f"{COLOR_RED}Error{COLOR_CLEAR}: Speech synthesis canceled: {cancellation_details.reason}",
              file=sys.stderr)
        if cancellation_details.reason == speechsdk.CancellationReason.Error:
            print(cancellation_details.error_details, file=sys.stderr)
        sys.exit(2)
    else:
        print(f"{COLOR_RED}Error{COLOR_CLEAR}: Unexpected result reason: {result.reason}", file=sys.stderr)
        sys.exit(3)



from ..quality import QUALITIES
from ..formats import get_available_formats


def list_qualities_and_formats():
    print('Available qualities:')
    for ext, info in QUALITIES.items():
        print(f"Qualities for {ext}:")
        for key, val in info.items():
            print(f"{key:2}: {val.name}")
    print()
    formats = get_available_formats()
    print("Available formats:")
    for format_ in formats:
        print(f'- {format_}')
